////////////////////////////////////////////////////////////////////////////////
///                           NTC_TEMPERATURE.C                              ///
///    Driver for temperature readout using NTC thermistors and PIC ADC      ///
///                                                                          ///
/// Copyright 2017-2018 Jaakko Kairus                                        ///
/// jaakko@kair.us                                                           ///
///                                                                          ///
/// You may freely use and modify this software for personal use.            ///
/// Commercial use of this software or parts of it is prohibited             ///
/// without written permission from author.                                  ///
/// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY              ///
///                                                                          ///
/// Three functions are provided:                                            ///
///                                                                          ///
/// read_ntc_sensors() must be called periodically to sample the ADC         ///
/// readings. The function executes very quickly, as it reads the previous   ///
/// result and starts a new conversion, so it suits well to be called from   ///
/// an interrupt. Oversampling is used, two separate sensor inputs are read  ///
/// Results are stored to in_sum and out_sum.                                ///
///                                                                          ///
/// lookup_32bit() uses integer arithmetic and lookup table to calculate     ///
/// temperature from oversampled sums acquired with read_ntc_sensors()       ///
///                                                                          ///
/// calculate_32bit() uses floating point arithmetic and calculates          ///
/// temperature using r25 and beta values. Allows easy way of using NTCs     ///
/// with mixed characteristics, with the expense of accuracy and code size.  ///
/// However if you already use floating point arithmetic, this can yield     ///
/// smaller code size because lookup table data is not needed.               ///
///                                                                          ///
/// For more information about projects which use this driver see project    ///  
/// web pages at                                                             ///
/// http://kair.us/projects/temp_controller/                                 ///
/// http://kair.us/projects/clock/v4/                                        ///
///                                                                          ///
/// Developed with CCS compiler version 5.079                                ///
///                                                                          ///
////////////////////////////////////////////////////////////////////////////////

// Uncomment the following line if you have NTC as pull-up to ADC positive reference
// and fixed resistor as pull-down to negative reference
// #define NTC_IS_PULLUP
// ADC channels used for NTCs
#ifndef NTC_CHANNEL_IN
#define NTC_CHANNEL_IN 11
#endif
#ifndef NTC_CHANNEL_OUT
#define NTC_CHANNEL_OUT 13
#endif
// Uncomment the following line if you are using lookup table based method.
// #define NTC_LOOKUP_TABLE
// If hysteresis is defined, a 0.025C hysteresis is used by lookup table method to reduce flickering
// of temperature reading if valus is between the tenths.
// #define NTC_USE_HYSTERESIS
// Pull-up / pull-down fixed resistor value in ohms
#define R_DIVIDER 10000.0
// ADC oversampling times. The lookup table is calculated for 512 oversampling times.
#define NTC_OVERSAMPLING 512
// Full scale of oversampled ADC reading.
//#define NTC_FULL_SCALE (2^getenv("ADC_RESOLUTION")*NTC_OVERSAMPLING)
#define NTC_FULL_SCALE 4096*NTC_OVERSAMPLING
#include <math.h>

// Lookup table for 10k B=3950 thermistor (eBay/Adafruit)
#ifdef NTC_LOOKUP_TABLE
unsigned int32 rom lookup[146] =   {2024131,2020502,2016523,2012178,2007530,2002472,1997097,1991289,1985184,1978669,
									1971724,1964505,1956780,1948734,1940297,1931500,1922389,1912706,1902791,1892552,
									1881905,1871141,1860025,1848556,1836733,1824511,1811980,1799007,1785678,1772012,
									1757917,1743381,1728519,1713199,1697542,1681463,1665017,1648083,1630807,1613157,
									1595081,1576120,1556927,1537165,1517187,1496937,1476142,1455232,1433916,1412479,
									1390564,1368468,1346292,1323866,1301269,1278592,1255936,1232704,1209656,1186930,
									1163840,1140422,1117632,1094210,1071650,1048576,1025810,1003176,980695,958386,
									936195,914325,892585,871177,849961,829071,808421,788068,767990,748328,
									728879,709871,691083,672747,654720,637045,619764,602815,586236,570063,
									554221,538742,523659,508882,494560,480476,466777,453487,440503,427977,
									415664,403715,392012,380707,369821,359085,348651,338681,328896,319302,
									310060,301184,292529,284105,275917,267975,260284,252853,245526,238634,
									231856,225196,218825,212580,206634,200821,195318,189870,184634,179545,
									174623,169835,165183,160670,156279,152031,147890,143878,139977,136189,
									132496,129067,125700,122414,119192,116053};
#endif
static unsigned int32 adc_sum=0,in_sum,out_sum;
static unsigned int16 k=0;
static unsigned int8 adc_selector=NTC_CHANNEL_IN, sums_ready=1;
static signed int16 in_temp, out_temp;
#ifdef NTC_USE_HYSTERESIS
	static signed int32 inhyst, outhyst;
#endif

// Function prototypes
int1 calculate_32bit(unsigned int32* adc_sum, signed int16* temperature, unsigned int16 ntc_r25, unsigned int16 ntc_beta);
void read_ntc_sensors();

#ifdef NTC_LOOKUP_TABLE
#ifdef NTC_USE_HYSTERESIS
void lookup_32bit(unsigned int32* adc_sum, signed int16* temperature, signed int32* temphyst)
#else
void lookup_32bit(unsigned int32* adc_sum, signed int16* temperature)
#endif

{
	signed int32 temp1,temp2;
	unsigned int32 ylempi,alempi;
	signed int16 new_temp;
	unsigned int16 j;

	for (j=0;j<146;j++)
	{
		if (*adc_sum>lookup[j])
			break;
	}
	
	ylempi = lookup[j];
	alempi = lookup[j-1];
	temp1=10*(alempi-*adc_sum);
	temp2=(alempi-ylempi);

#ifdef NTC_USE_HYSTERESIS
	new_temp = (temp1+*temphyst)/temp2+ 10*(j-41); // -41 because our lookup table begins from -40

	if (new_temp > *temperature)
	{
		*temphyst = temp2/4;
		*temperature = new_temp;
	}	
	if (new_temp < *temperature)
	{
		*temphyst = -temp2/4;
		*temperature = new_temp;
	}	

	new_temp = temp1%temp2;
	temp2=temp2/3;
	temp1=2*temp2;
	if (temp2 < new_temp && new_temp < temp1)	// zero hysteresis if not close to flicker
		*temphyst = 0;
#else
	new_temp = temp1/temp2+ 10*(j-41); // -41 because our lookup table begins from -40
	*temperature = new_temp;
#endif
		
}
#endif		// end NTC_LOOKUP_TABLE

int1 calculate_32bit(unsigned int32* adc_sum, signed int16* temperature, unsigned int16 ntc_r25, unsigned int16 ntc_beta)
{
	if (*adc_sum >= ((NTC_FULL_SCALE-1)&0xFFFF0000))		// 2031616 in case of 512 times oversampling of 12 bit ADC.
		return 0;											// May need to be changed if NTC_FULL_SCALE changes radically
	float ntc_r25_f = (float)ntc_r25;
	float ntc_beta_f = (float)ntc_beta;
	float ntc_beta_per_r25 = ntc_beta_f / 298.15;
	float l, t ,r;
	
	#ifdef NTC_IS_PULLUP
		r = NTC_FULL_SCALE * (float)R_DIVIDER;
		r -= *adc_sum * (float)R_DIVIDER; 
		r /= *adc_sum;
	#else	// NTC is pulldown and fixed resistor is pullup
		r = *adc_sum * (float)R_DIVIDER;
		r /= (NTC_FULL_SCALE - *adc_sum);
	#endif
	  
	l = log(r/ntc_r25_f);
	
	t = ntc_beta_f / (l + ntc_beta_per_r25);  //we now have temp in kelvin

	t -= (float)273.15;
	t *= (float)10;   // in tenths
	
	*temperature = (signed int16)t;
	return 1;
}

void read_ntc_sensors() {
	if(adc_done() && !(sums_ready)) {
		if (k!=0)				// discard the first conversion after mux change
			adc_sum += read_adc(ADC_READ_ONLY);
		
		if (k<NTC_OVERSAMPLING+1) {
			k++;
		}
		else {
			if (adc_selector==NTC_CHANNEL_IN) {
				adc_selector=NTC_CHANNEL_OUT;
				in_sum = adc_sum;
			}
			else {
				adc_selector=NTC_CHANNEL_IN;
				out_sum = adc_sum;
				sums_ready = 1;
			}		
					
			set_adc_channel(adc_selector);
			k=0;
			adc_sum = 0;
		}
		read_adc(ADC_START_ONLY);
	}
}	
